using Adaptation.Shared;
using Adaptation.Shared.Methods;
using log4net;
using System;
using System.Collections.Generic;
using System.Data;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text.Json;

namespace Adaptation.FileHandlers.jpeg;

public class ProcessData : IProcessData
{

    private readonly List<object> _Details;

    private readonly ILog _Log;

    public string JobID { get; set; }
    public string MesEntity { get; set; }

    public int Red { get; set; }
    public int Green { get; set; }
    public string Text { get; set; }
    public int TotalDelta { get; set; }
    public string ClosestMatchFileName { get; set; }

    List<object> Shared.Properties.IProcessData.Details => _Details;

    public ProcessData(IFileRead fileRead, Logistics logistics, List<FileInfo> fileInfoCollection, int startX, int startY, int endX, int endY, List<(string, Color[])> colorCollections)
    {
        if (logistics is null)
        { }
        JobID = logistics.JobID;
        fileInfoCollection.Clear();
        _Details = new List<object>();
        MesEntity = logistics.MesEntity;
        _Log = LogManager.GetLogger(typeof(ProcessData));
        Parse(fileRead, fileInfoCollection, startX, startY, endX, endY, colorCollections);
    }

    private static string Get(string value, bool useSplitForMID)
    {
        string result = value;
        if (useSplitForMID)
        {
            if (result.IndexOf(".") > -1)
                result = result.Split('.')[0].Trim();
            if (result.IndexOf("_") > -1)
                result = result.Split('_')[0].Trim();
            if (result.IndexOf("-") > -1)
                result = result.Split('-')[0].Trim();
        }
        result = string.Concat(result.Substring(0, 1).ToUpper(), result.Substring(1).ToLower());
        return result;
    }

    string IProcessData.GetCurrentReactor(IFileRead fileRead, Logistics logistics, Dictionary<string, string> reactors)
    {
        string result = string.Empty;
        string filePrefixAsMID;
        foreach (KeyValuePair<string, string> keyValuePair in reactors)
        {
            foreach (string filePrefix in keyValuePair.Value.Split('|'))
            {
                filePrefixAsMID = Get(filePrefix, useSplitForMID: true);
                if (logistics.MID.StartsWith(filePrefix) || logistics.MID.StartsWith(filePrefixAsMID))
                {
                    result = keyValuePair.Key;
                    break;
                }
            }
        }
        if (string.IsNullOrEmpty(result) && reactors.Count == 1)
            result = reactors.ElementAt(0).Key;
        return result;
    }

    Tuple<string, Test[], JsonElement[], List<FileInfo>> IProcessData.GetResults(IFileRead fileRead, Logistics logistics, List<FileInfo> fileInfoCollection)
    {
        Tuple<string, Test[], JsonElement[], List<FileInfo>> results;
        List<Test> tests = new();
        foreach (object item in _Details)
            tests.Add(Test.CDE);
        List<IDescription> descriptions = fileRead.GetDescriptions(fileRead, tests, this);
        if (tests.Count != descriptions.Count)
            throw new Exception();
        for (int i = 0; i < tests.Count; i++)
        {
            if (descriptions[i] is not Description description)
                throw new Exception();
            if (description.Test != (int)tests[i])
                throw new Exception();
        }
        List<Description> fileReadDescriptions = (from l in descriptions select (Description)l).ToList();
        string json = JsonSerializer.Serialize(fileReadDescriptions, fileReadDescriptions.GetType());
        JsonElement[] jsonElements = JsonSerializer.Deserialize<JsonElement[]>(json);
        results = new Tuple<string, Test[], JsonElement[], List<FileInfo>>(logistics.Logistics1[0], tests.ToArray(), jsonElements, fileInfoCollection);
        return results;
    }

#nullable enable
#pragma warning disable CA1416

    private static Color[] Get(string reportFullPath, int startX, int startY, int endX, int endY)
    {
        Color color;
        List<Color> colors = new();
        using Bitmap? bitmap = Image.FromFile(reportFullPath) as Bitmap;
        if (bitmap is null)
            throw new Exception($"Couldn't load image from <{reportFullPath}>");
        for (int x = startX; x < endX; x++)
        {
            for (int y = startY; y < endY; y++)
            {
                color = bitmap.GetPixel(x, y);
                colors.Add(color);
            }
        }
        return colors.ToArray();
    }

    private static System.Drawing.Imaging.ImageFormat Get(string extension)
    {
        System.Drawing.Imaging.ImageFormat imageFormat = extension switch
        {
            ".bmp" => System.Drawing.Imaging.ImageFormat.Bmp,
            ".gif" => System.Drawing.Imaging.ImageFormat.Gif,
            ".jpeg" => System.Drawing.Imaging.ImageFormat.Jpeg,
            ".jpg" => System.Drawing.Imaging.ImageFormat.Jpeg,
            ".png" => System.Drawing.Imaging.ImageFormat.Png,
            ".tiff" => System.Drawing.Imaging.ImageFormat.Tiff,
            _ => throw new Exception("Extension not mapped"),
        };
        return imageFormat;
    }

    private static (Color[], int, int) Get(IFileRead fileRead, List<FileInfo> fileInfoCollection, string extension, int startX, int startY, int endX, int endY)
    {
        Color color;
        List<Color> colors = new();
        MemoryStream memoryStream = new();
        Bitmap selectedBitmap = new(endX - startX, endY - startY);
        System.Drawing.Imaging.ImageFormat imageFormat = Get(extension);
        using Bitmap? bitmap = Image.FromFile(fileRead.ReportFullPath) as Bitmap;
        string saveFileName = Path.ChangeExtension(fileRead.ReportFullPath, extension);
        if (bitmap is null)
            throw new Exception($"Couldn't load image from <{fileRead.ReportFullPath}>");
        for (int x = startX; x < endX; x++)
        {
            for (int y = startY; y < endY; y++)
            {
                color = bitmap.GetPixel(x, y);
                colors.Add(color);
                selectedBitmap.SetPixel(x - startX, y - startY, color);
            }
        }
        selectedBitmap.Save(memoryStream, imageFormat);
        if (!fileRead.IsEAFHosted)
        {
            fileInfoCollection.Add(new FileInfo(saveFileName));
            SaveToFile(extension, saveFileName, memoryStream);
        }
        return new(colors.ToArray(), endX - startX, endY - startY);
    }

    private static string Get(IFileRead fileRead, string extension, string extra)
    {
        string result;
        string? directoryName = Path.GetDirectoryName(fileRead.ReportFullPath);
        if (string.IsNullOrEmpty(directoryName))
            throw new Exception("Couldn't get directoryName!");
        string? fileNameWithoutExtension = Path.GetFileNameWithoutExtension(fileRead.ReportFullPath);
        if (string.IsNullOrEmpty(fileNameWithoutExtension))
            throw new Exception("Couldn't get fileNameWithoutExtension!");
        result = Path.Combine(directoryName, $"{fileNameWithoutExtension} - {extra}{extension}");
        return result;
    }

    private static void SaveToFile(string extension, string saveFileName, MemoryStream memoryStream)
    {
        System.Drawing.Imaging.ImageFormat imageFormat = extension switch
        {
            ".bmp" => System.Drawing.Imaging.ImageFormat.Bmp,
            ".gif" => System.Drawing.Imaging.ImageFormat.Gif,
            ".jpeg" => System.Drawing.Imaging.ImageFormat.Jpeg,
            ".jpg" => System.Drawing.Imaging.ImageFormat.Jpeg,
            ".png" => System.Drawing.Imaging.ImageFormat.Png,
            ".tiff" => System.Drawing.Imaging.ImageFormat.Tiff,
            _ => throw new Exception("Extension not mapped"),
        };
        using Bitmap bitmap = new(memoryStream);
        bitmap.Save(saveFileName, imageFormat);
    }

#pragma warning restore CA1416

    private static void SaveText(int red, int green, string closestMatchFileNameEnding, string textFileName, List<(string File, int TotalDelta)> totalDeltaCollection)
    {
        string format = "00000";
        List<string> lines = new() { red.ToString(format), green.ToString(format), closestMatchFileNameEnding };
        foreach ((string file, int totalDelta) in totalDeltaCollection)
        {
            lines.Add(file);
            lines.Add(totalDelta.ToString(format));
        }
        File.WriteAllLines(textFileName, lines);
    }

    internal static List<(string, Color[])> GetColorCollections(int startX, int startY, int endX, int endY, string masterImageDirectory)
    {
        List<(string, Color[])> results = new();
        string[] files = Directory.GetFiles(masterImageDirectory, "*.jpeg", SearchOption.TopDirectoryOnly);
        foreach (string file in files)
            results.Add(new(file, Get(file, startX, startY, endX, endY)));
        return results;
    }

    private void Parse(IFileRead fileRead, List<FileInfo> fileInfoCollection, int startX, int startY, int endX, int endY, List<(string, Color[])> colorCollections)
    {
        Red = 0;
        Green = 0;
        int delta;
        Color color;
        int totalDelta;
        const int thresHold = 70;
        const string extension = ".tiff";
        List<(string File, int TotalDelta)> totalDeltaCollection = new();
        (Color[] sourceColors, int width, int height) = Get(fileRead, fileInfoCollection, extension, startX, startY, endX, endY);
        foreach ((string file, Color[] colors) in colorCollections)
        {
            totalDelta = 0;
            if (colors.Length != sourceColors.Length)
                continue;
            for (int i = 0; i < sourceColors.Length; i++)
            {
                color = sourceColors[i];
                if (color.R > thresHold)
                    Red += 1;
                if (color.G > thresHold)
                    Green += 1;
                delta = color.R - colors[i].R;
                if (delta > 0)
                    totalDelta += delta;
                else
                    totalDelta += delta * -1;
                delta = color.G - colors[i].G;
                if (delta > 0)
                    totalDelta += delta;
                else
                    totalDelta += delta * -1;
                delta = color.B - colors[i].B;
                if (delta > 0)
                    totalDelta += delta;
                else
                    totalDelta += delta * -1;
            }
            totalDeltaCollection.Add(new(file, totalDelta));
        }
        totalDeltaCollection = (from l in totalDeltaCollection orderby l.TotalDelta select l).ToList();
        ClosestMatchFileName = Path.GetFileNameWithoutExtension(totalDeltaCollection[0].File);
        string[] closestMatchFileNameSplit = ClosestMatchFileName.Split('-');
        Text = closestMatchFileNameSplit.Last().TrimStart();
        TotalDelta = totalDeltaCollection[0].TotalDelta;
        string textFileName = Get(fileRead, ".txt", $"{TotalDelta} - {Text}");
        if (!fileRead.IsEAFHosted)
        {
            _Log.Debug(textFileName);
            fileInfoCollection.Add(new FileInfo(textFileName));
            SaveText(Red, Green, Text, textFileName, totalDeltaCollection);
        }
        _Details.Add(Text);
    }

}